from __future__ import annotations

import sys

from uvpipx.internal_libs.Logger import Logger, LogMode, get_logger
from uvpipx.platform.win import get_env_variable, set_env_variable
from uvpipx.uvpipx_core import UvPipxVenv
from uvpipx.uvpipx_uv import uv_get_version
from uvpipx.uvpipx_venv_factory import path_link_from_model
from uvpipx.uvpipx_venv_load import uvpipx_load_venv
from uvpipx.UvPipxModels import UvPipxModel
from uvpipx.version import show_version

__author__ = "Gaëtan Montury"
__copyright__ = "Copyright (c) 2024-2025 Gaëtan Montury"
__license__ = """GNU GENERAL PUBLIC LICENSE refer to file LICENSE in repo"""
__version__ = "0.8.1"  # to bump
__maintainer__ = "Gaëtan Montury"
__email__ = "#"
__status__ = "Development"


import os
from pathlib import Path
from typing import List, Union

import uvpipx.platform
from uvpipx import config
from uvpipx.internal_libs.misc import shell_run


def ensurepath(just_check: bool = False) -> str:
    if uvpipx.platform.sys_platform == "win":
        return ensurepath_win(just_check)

    return ensurepath_unix(just_check)


def ensurepath_unix(just_check: bool = False) -> str:
    logger = Logger(log_mode=LogMode.BUFFER) if just_check else get_logger("ensurepath")

    logger.log_info(f"Search {config.uvpipx_local_bin} in PATH\n")

    path_os_setted = os.environ["PATH"].split(":")
    if str(config.uvpipx_local_bin) in path_os_setted:
        logger.log_info("🟢 Configuration of PATH already OK")
        return "OK"

    profile_path = Path(os.environ["HOME"]) / ".profile"

    profile_content = Path(profile_path).read_text()
    if "# Added by uvpipx" in profile_content:
        logger.log_info(f"""🟣  Configuration already in {profile_path}

⚠️  To use without restart the shell, launch 
export PATH=$PATH:{config.uvpipx_local_bin}
""")
        return "OK/RESTART_NEED"

    if just_check:
        return "KO"

    line_to_add = "\n" if Path(profile_path).read_text()[-1] != "\n" else ""
    line_to_add += f"""# Added by uvpipx
export PATH=$PATH:{config.uvpipx_local_bin}\n"""

    with profile_path.open("a") as file:
        file.write(line_to_add)

    logger.log_info(f"""⚙️  Configuration added to {profile_path}

⚠️  To use without restart the shell, launch 
export PATH=$PATH:{config.uvpipx_local_bin}
""")

    return "OK/RESTART_NEED"


def ensurepath_win(just_check: bool = False) -> str:
    logger = Logger(log_mode=LogMode.BUFFER) if just_check else get_logger("ensurepath")

    logger.log_info(f"Search {config.uvpipx_local_bin} in PATH\n")

    path_os_setted = os.environ["PATH"].split(";")
    if str(config.uvpipx_local_bin) in path_os_setted:
        logger.log_info("🟢 Configuration of PATH already OK")
        return "OK"

    path_reg_level = None
    path_reg_user = get_env_variable("PATH")
    if str(config.uvpipx_local_bin) in path_reg_user.split(";"):
        path_reg_level = "user"

    if path_reg_level is None:
        path_reg_system = get_env_variable("PATH", system=True)
        if str(config.uvpipx_local_bin) in path_reg_system.split(";"):
            path_reg_level = "system"

    if path_reg_level:
        logger.log_info(f"""🟣  Configuration of PATH already in registry {path_reg_level} but not in shell.
                        
⚠️ you should restart the shell to use the new configuration.""")
        return "OK/RESTART_NEED"

    set_env_variable("PATH", f"{config.uvpipx_local_bin};{path_reg_user}")
    logger.log_info("""⚙️  Configuration added to PATH in user environnement vars.

⚠️ you should restart the shell to use the new configuration.""")

    return "OK/RESTART_NEED"


def _info(uvpipx: UvPipxModel, venv: UvPipxVenv) -> str:
    def get_version(pck_name: str, stdout_line: List[str]) -> str:
        vers = next(
            (line for line in stdout_line if pck_name in line),
            None,
        )
        if vers is None:
            vers = f"{pck_name} Unknown version"

        return vers

    rc, stdout, stderr = shell_run("uv pip freeze", cwd=venv.venv_path)
    stdout_line = stdout.split("\n")
    main_vers = get_version(uvpipx.main_package.package_name, stdout_line)

    injected_vers = []
    for pkg_injected in list(uvpipx.injected_packages.keys()):
        injected_vers.append(get_version(pkg_injected, stdout_line))
    injected_vers = sorted(injected_vers)

    path_links = (
        [path_link_from_model(uvpipx.exposed, m) for m in uvpipx.exposed.apps.values()]
        if uvpipx.exposed and uvpipx.exposed.apps
        else []
    )
    bins = "\n".join(f"   ✅ {app_bin.show_name_with_link()}" for app_bin in path_links)

    # TODO add show info about install set in advanced view

    if not bins:
        bins = "   ❌ Nothing exposed"

    output = f""" 📦 {uvpipx.main_package.package_name} ({main_vers}) in venv {uvpipx.venv.uvpipx_dir}"""
    if injected_vers:
        output_inject = "\n".join(f"""   📦 {pkg_ver}""" for pkg_ver in injected_vers)
        output += f"""
 🎯 Injected packages
{output_inject}"""

    output += f"""
 🎯 Exposed program
{bins}"""

    return output


def _info_pkg(
    package_name: str,
    *,
    name_override: Union[None, str] = None,
    get_venv: bool = False,
) -> str:
    uvpipx, venv = uvpipx_load_venv(package_name, name_override)

    return str(venv.venv_path / ".venv") if get_venv else _info(uvpipx, venv)


def info(
    package_name: str,
    *,
    name_override: Union[None, str] = None,
    get_venv: bool = False,
) -> None:
    logger = get_logger("info")

    info = _info_pkg(package_name, name_override=name_override, get_venv=get_venv)
    logger.log_info(info)


def uvpipx_list() -> None:
    infos = f"""📍 uvpipx venvs are in {config.uvpipx_venvs}
💻 apps are exposed at {config.uvpipx_local_bin} 

"""
    logger = get_logger("uvpipx_list")

    show_version()

    nb = 0
    if config.uvpipx_venvs.exists():
        for pck_venv in config.uvpipx_venvs.iterdir():
            if pck_venv.name.startswith("."):
                continue
            infos += _info_pkg(pck_venv.name)  # this is a tricky way to get the name
            infos += "\n\n"
            nb += 1

    if nb == 0:
        infos += "⭕ No uvpipx package installed!"

    logger.log_info(infos)


def uvpipx_show_config() -> None:
    logger = get_logger("uvpipx_show_config")

    show_version()

    uv_version = uv_get_version()
    py_version = f"{sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}"

    logger.log_info("🔍 Uvpipx configuration:")

    logger.log_info(f"💻 platform = {uvpipx.platform.sys_platform}")
    logger.log_info(f"💻 python version = {py_version}")
    logger.log_info(f"💻 uv version = {uv_version}")
    logger.log_info(f"🏠 user home = {config.home}")

    logger.log_info(f"\n🌳 uvpipx home = {config.uvpipx_home}")
    logger.log_info("    Path to the main directory of uvpipx.")
    if uvpipx.platform.sys_platform == "win":
        logger.log_info(
            "    🎚️  Defined by the UVPIPX_HOME environment variable or defaults to $HOME/.local/uv-pipx",
        )
    else:
        logger.log_info(
            "    🎚️  Defined by the UVPIPX_HOME environment variable or defaults to ~/.local/uv-pipx",
        )

    logger.log_info(f"\n🌿 uvpipx venvs = {config.uvpipx_venvs}")
    logger.log_info("    Path to the directory of uvpipx virtual environments.")
    logger.log_info(
        "    🎚️  Defined by the UVPIPX_LOCAL_VENVS environment variable or defaults to $UVPIPX_HOME/venvs",
    )

    logger.log_info(f"\n📁 exposing bin directory = {config.uvpipx_local_bin}")
    logger.log_info("    Default path for exposed executables.")
    if uvpipx.platform.sys_platform == "win":
        logger.log_info(
            """    Default to %HOME%/.local/bin""",
        )
    else:
        logger.log_info(
            """    Default:
            ~/.local/bin for normal users or
            /usr/local/bin for root.""",
        )
    logger.log_info(
        "    🎚️  Can be defined by the UVPIPX_BIN_DIR environment variable",
    )
    # logger.log_info(
    #     "    Note: The value for Windows is not defined in the provided code.",
    # )



from __future__ import annotations

import json
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Tuple, Union

import uvpipx
import uvpipx.platform
from uvpipx import config
from uvpipx.internal_libs.Logger import Logger, get_logger
from uvpipx.uvpipx_core import UvPipxVenv
from uvpipx.uvpipx_venv_factory import path_link_factory
from uvpipx.uvpipx_venv_load import uvpipx_load_venv
from uvpipx.UvPipxModels import UvPipxExposedModel, UvPipxVenvExposeAppModel


@dataclass
class ExposeApps:
    venv: UvPipxVenv
    logger: Union[None, Logger] = None
    prev_exposed: Union[None, UvPipxExposedModel] = None

    def __post_init__(self) -> None:
        self.logger_ = self.logger or get_logger()

    def set_prev_exposed(self, prev_exposed: Union[None, UvPipxExposedModel]) -> None:
        self.prev_exposed = prev_exposed

    def get_apps_list(
        self,
        expose_app_rules: List[str],
        main_package_name: str,
    ) -> Tuple[List[Path], Dict[str, str]]:
        # expose_apps_list = [
        #     Path(e) if isinstance(e, str) else e for e in expose_app_rules
        # ]
        renamed_apps = {}

        if not expose_app_rules:
            return ([], {})

        expose_apps_list = None
        expose_fallback = None
        if expose_app_rules == ["__eponym__"]:
            if self.venv.venv_bin(main_package_name).exists():
                expose_apps_list = [self.venv.venv_bin(main_package_name)]
            else:
                expose_fallback = "__main__"
                self.logger_.log_warn(
                    f" ⚠️  no find eponym app {self.venv.venv_bin(main_package_name)}. will fallback to app of main package",
                )

        if expose_app_rules == ["__main__"] or expose_fallback == "__main__":
            self.venv.update_metadata()

            with (self.venv.venv_path / "pip_metadata.json").open("r") as infile:
                metadata_dict = json.load(infile)

            console_scripts = metadata_dict["console_scripts"].get(
                main_package_name,
                [],
            )
            if console_scripts:
                expose_apps_list = [
                    self.venv.venv_bin(main_bin)
                    for main_bin in console_scripts
                    if not re.match(r"^(python|pip)(\d+(\.\d+)?)?$", main_bin)
                ]
                expose_fallback = None
            else:
                expose_fallback = "__eponym__"
                self.logger_.log_warn(
                    " ⚠️  cannot find metadata of package. fallback to eponym app",
                )

        if expose_fallback == "__eponym__":
            if self.venv.venv_bin(main_package_name, fail_if_notexist=False).exists():
                self.logger_.log_info(" 🔰 fallback to eponym app is OK ")
                expose_fallback = None
                expose_apps_list = [self.venv.venv_bin(main_package_name)]
            else:
                expose_fallback = "__all__"
                self.logger_.log_warn(
                    f" ⚠️  fallback also failed: unable to find eponym app {main_package_name}. will fallback to all app in venv",
                )

        if expose_app_rules == ["__all__"] or expose_fallback == "__all__":
            if uvpipx.platform.sys_platform == "win":
                expose_apps_list = self.venv.venv_bins(
                    regex_to_exclude=[r"^(python|pythonw|pip)(\d+(\.\d+)?)?(\.exe)?$"],
                )
            else:
                expose_apps_list = self.venv.venv_bins(
                    regex_to_exclude=[r"^(python|pip)(\d+(\.\d+)?)?$"],
                )
        elif not expose_apps_list:
            renamed_apps = {
                key: value for item in expose_app_rules if ":" in item for key, value in [item.split(":", 1)]
            }
            expose_apps_list = [self.venv.venv_bin(item.split(":", 1)[0]) for item in expose_app_rules]

        return expose_apps_list, renamed_apps

    def get_apps_to_remove(
        self,
        future_exposed_apps: Dict[str, UvPipxVenvExposeAppModel],
    ) -> Dict[str, Path]:
        if self.prev_exposed is None:
            return {}

        prev_exposed_apps = {k: Path(app.exposed_app_path) for k, app in self.prev_exposed.apps.items()}
        # remove pre_exposed_apps that are not in future_exposed_apps
        return {k: v for k, v in prev_exposed_apps.items() if k not in future_exposed_apps}

    def add_exposing(
        self,
        exposing_apps: List[Path],
        renamed_apps: Dict[str, str],
        pkgs_sets: List[str],
    ) -> Dict[str, UvPipxVenvExposeAppModel]:
        config.uvpipx_local_bin.mkdir(parents=True, exist_ok=True)
        exposed_apps = {}

        for app_bin in exposing_apps:
            pl = path_link_factory(
                app_bin,
                config.uvpipx_local_bin,
                rename_local_bin=renamed_apps.get(app_bin.name),
            )
            if not pl.link_exists():
                self.logger_.log_info(
                    f" 🎯 Exposing program {pl.show_name_with_link()}",
                )
                pl.link()
                exposed_apps[app_bin.name] = UvPipxVenvExposeAppModel(
                    app_bin.name,
                    str(pl.link_path),
                    pkgs_sets,
                )
            elif pl.is_valid():
                self.logger_.log_info(
                    f" 🔵 Already exposed to current uvpipx venv program {pl.show_name_with_link()}",
                )
                exposed_apps[app_bin.name] = UvPipxVenvExposeAppModel(
                    app_bin.name,
                    str(pl.link_path),
                    pkgs_sets,
                )
            else:
                if pl.link_path:
                    self.logger_.log_warn(
                        f" ❌ Not exposing program {app_bin.name}; {pl.link_path} already exist to {pl.link_path.resolve()}!",
                    )
                else:
                    msg = "link_path is None"
                    raise ValueError(msg)

        return exposed_apps

    def remove_exposing(
        self,
        remove_apps: Dict[str, Path],
    ) -> None:
        for app_bin, link_path in remove_apps.items():
            if link_path.exists():
                self.logger_.log_info(
                    f" 🗑️  Removing exposing {link_path} -> {app_bin} ",
                )
                link_path.unlink()

    def expose(
        self,
        main_package_name: str,
        expose_app_rules: List[str],
        pkgs_sets: List[str],
    ) -> Dict[str, UvPipxVenvExposeAppModel]:
        exposing_apps, renamed_apps = self.get_apps_list(
            expose_app_rules,
            main_package_name,
        )

        exposed_apps = self.add_exposing(exposing_apps, renamed_apps, pkgs_sets)
        remove_apps = self.get_apps_to_remove(exposed_apps)
        self.remove_exposing(remove_apps)

        return exposed_apps

    # TODO add function to clean old link in local bin


def expose(package_name: str, expose_rule_names: List[str]) -> None:
    logger = get_logger("expose")

    logger.log_info(f"🔭  Exposing apps of {package_name}\n")

    expose_rule_names_ = expose_rule_names

    uvpipx_model, venv = uvpipx_load_venv(
        package_name,
    )
    expo_app = ExposeApps(venv, logger)
    expo_app.set_prev_exposed(uvpipx_model.exposed if uvpipx_model else None)
    package_name = uvpipx_model.main_package.package_name

    uvpipx_model.exposed.apps = expo_app.expose(
        package_name,
        expose_rule_names_,
        [package_name],
    )
    uvpipx_model.save_json("uvpipx.json")

    logger.log_info(
        f" 🟢 uvpipx venv {uvpipx_model.venv.name()} with {package_name} ready",
    )


def expose_all(expose_rule_names: List[str]) -> None:
    logger = get_logger("expose_all")

    infos = ""

    nb = 0
    if config.uvpipx_venvs.exists():
        for pck_venv in config.uvpipx_venvs.iterdir():
            if nb > 0:
                logger.log_info("")
            expose(
                pck_venv.name,
                expose_rule_names,
            )  # This is a tricky way to get the name
            logger.log_info(" ----------------")
            nb += 1

    if nb == 0:
        infos += "⭕ No uvpipx package installed!"

    if infos:
        logger.log_info(infos)



from __future__ import annotations

import os
import re
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Tuple, Union

import uvpipx.platform
from uvpipx import config
from uvpipx.internal_libs.misc import file_md5, find_executable, shell_run
from uvpipx.req_spec import Requirement


@dataclass
class UvPipxVenv:
    venv_path: Path

    def exists(self) -> bool:
        return (self.venv_path / ".venv").exists()

    def create_venv_if_need(self) -> bool:
        self.venv_path.mkdir(exist_ok=True, parents=True)

        if not self.exists():
            rc, std_o, std_e = shell_run(
                f"uv venv {self.venv_path / '.venv'}",
                raise_on_error=True,
            )
            return True
        return False

    def freeze(self) -> str:
        rc, stdout, stderr = shell_run("uv pip freeze", cwd=self.venv_path)
        return stdout  # if isinstance(stdout, str) else stdout.decode("utf-8")

    def installed_package(self) -> list[str]:
        return self.freeze().rstrip().split("\n")

    def installed_package_as_req(self) -> list[Requirement]:
        req = self.freeze().rstrip().split("\n")
        req_tuple = [Requirement.from_str(r) for r in req]

        return req_tuple

    def install(
        self,
        packages_name_spec: List[str],
        allow_upgrade: bool = False,
    ) -> str:
        opt = " --upgrade" if allow_upgrade else ""
        install_pkgs = " ".join(packages_name_spec)
        rc, stdout, stderr = shell_run(
            f"uv pip install{opt} {install_pkgs}",
            cwd=self.venv_path,
        )
        return stdout  # if isinstance(stdout, str) else stdout.decode("utf-8")

    def uninstall(self, package_name_spec: str) -> str:
        rc, stdout, stderr = shell_run(
            f"uv pip uninstall {package_name_spec}",
            cwd=self.venv_path,
        )
        return stdout  # if isinstance(stdout, str) else stdout.decode("utf-8")

    def venv_bin_dir(self) -> Path:
        return self.venv_path / ".venv" / uvpipx.platform.venv_bin_dir

    def venv_bins(self, regex_to_exclude: list[str]) -> list[Path]:
        re_comp_to_exclude = [re.compile(r) for r in regex_to_exclude]
        all_bins = find_executable(self.venv_bin_dir())
        return [s for s in all_bins if not any(r.search(s.name) for r in re_comp_to_exclude)]

    def venv_bin(self, name: str, fail_if_notexist: bool = True) -> Path:
        path_ = self.venv_bin_dir() / (name + uvpipx.platform.bin_ext)

        if fail_if_notexist and not path_.exists():
            msg = f"🔴 {name} not exist (path {path_})"
            raise RuntimeError(msg)

        return path_

    def run_in_venv(
        self,
        cmdline: str,
        *,
        cwd: Union[None, Path] = None,
        env: Union[None, Dict[str, str]] = None,
    ) -> Tuple[int, str, str]:
        rc, stdo, stde = shell_run(cmdline, cwd=cwd, env=env)
        if rc != 0:
            msg = f"🔴 Command failed with return code {rc} {stdo} {stde}"
            raise RuntimeError(msg)

        return rc, stdo, stde

    def update_metadata(self, if_not_exist: bool = True) -> None:
        pip_metadata = self.venv_path / "pip_metadata.json"
        if pip_metadata.exists() and if_not_exist:
            return

        uvpipx_console_scripts = config.uvpipx_self_dir / "uvpipx/uvpipx_console_scripts.py"
        python_venv_bin = self.venv_bin_dir() / "python"
        self.run_in_venv(
            f"{python_venv_bin} {uvpipx_console_scripts} {self.venv_path} {pip_metadata}",
        )


@dataclass
class PathLink:
    local_path: Path
    link_path: Union[Path, None] = None

    def exists(self) -> bool:
        return self.local_path.exists()

    def link_exists(self) -> bool:
        if self.link_path is None:
            msg = "Link not defined"
            raise RuntimeError(msg)

        return self.link_path.exists()

    def is_valid(self) -> bool:
        if self.link_path is None:
            msg = "link_path is None"
            raise ValueError(msg)

        if uvpipx.platform.sys_platform == "win":
            return self.link_path.exists() and file_md5(self.link_path) == file_md5(
                self.local_path,
            )

        if self.link_path.exists():
            return self.link_path.is_symlink() and (self.local_path == self.link_path.resolve())

        return False

    def link(self) -> None:
        if self.link_path is None:
            msg = "link_path is None"
            raise ValueError(msg)

        if uvpipx.platform.sys_platform == "win":
            shutil.copy2(self.local_path, self.link_path)
        else:
            os.symlink(self.local_path, self.link_path)

    def unlink(self, if_valid: bool = True) -> None:
        if self.link_path is None:
            msg = "link_path is None"
            raise ValueError(msg)

        if (if_valid and self.is_valid()) or (not if_valid and self.exists()):
            self.link_path.unlink()
        # TODO maybe is not valid we shoud explain this

    def show_name_with_link(self) -> str:
        if self.link_path is None:
            msg = "link_path is None"
            raise ValueError(msg)

        return (
            self.link_path.name
            if self.link_path.name == self.local_path.name
            else f" {self.local_path.name} -> {self.link_path.name}"
        )



import os


def set_env_variable(name: str, value: str, system=False):
    import ctypes
    import winreg

    try:
        root_key = winreg.HKEY_LOCAL_MACHINE if system else winreg.HKEY_CURRENT_USER
        subkey = r"SYSTEM\CurrentControlSet\Control\Session Manager\Environment" if system else "Environment"

        key = winreg.OpenKey(root_key, subkey, 0, winreg.KEY_ALL_ACCESS)
        winreg.SetValueEx(key, name, 0, winreg.REG_EXPAND_SZ, value)
        winreg.CloseKey(key)

        ctypes.windll.user32.SendMessageW(65535, 0x001A, 0, "Environment")

        os.environ[name] = value
    except Exception as e:
        print(f"Error while setting environment variable: {e}")


def get_env_variable(name: str, system=False) -> str:
    """
    Lit une variable d'environnement.

    :param name: Le nom de la variable d'environnement
    :param system: Si True, lit la variable au niveau système, sinon au niveau utilisateur
    :return: La valeur de la variable d'environnement ou None si elle n'existe pas
    """
    import winreg

    try:
        root_key = winreg.HKEY_LOCAL_MACHINE if system else winreg.HKEY_CURRENT_USER
        subkey = r"SYSTEM\CurrentControlSet\Control\Session Manager\Environment" if system else "Environment"

        key = winreg.OpenKey(root_key, subkey, 0, winreg.KEY_READ)
        value, _ = winreg.QueryValueEx(key, name)
        winreg.CloseKey(key)
        return value
    except WindowsError:
        return None


# Exemple d'utilisation
if __name__ == "__main__":
    # var_name = input("Entrez le nom de la variable d'environnement : ")
    # var_value = input("Entrez la valeur de la variable d'environnement : ")
    # set_env_variable(var_name, var_value)
    pat = get_env_variable("PATH")
    print(pat)
    pass



from __future__ import annotations

from uvpipx.internal_libs.misc import shell_run


def uv_get_version() -> str:
    rc, stdout, stderr = shell_run("uv --version")
    return stdout.strip().removeprefix("uv").strip()
